"""Common set of utility functions and constants used across the project."""
from __future__ import annotations

import hashlib
import json
import logging
import operator
from contextlib import contextmanager
from datetime import datetime
from functools import partial, reduce
from pathlib import Path

import numpy as np


def is_valid_number(num) -> bool:
    """Check if the given number is a valid numerical value and not NaN."""
    return isinstance(num, (float, int, np.number)) and not np.isnan(num)


def safe_division(a: float, b: float, *, worst_result: float):
    """Try to divide the given arguments and return `worst_result` if unsuccessful."""
    if b == 0 or not is_valid_number(a) or not is_valid_number(b):
        logging.debug(
            f"Using `worst_result={worst_result}` in place of the following "
            f"division: {a} / {b}")
        return worst_result
    else:
        return a / b


def join_dictionaries(*dicts) -> dict:
    """Joins a sequence of dictionaries into a single dictionary."""
    return reduce(operator.or_, dicts)


def get_current_timestamp() -> str:
    """Return a timestamp representing the current time up to the second."""
    return datetime.now().strftime("%Y.%m.%d-%H.%M.%S")


def get_current_date() -> str:
    """Return a timestamp representing the current time up to the second."""
    return datetime.now().strftime("%Y-%m-%d")


def hash_dict(d: dict, length: int = 8) -> str:
    """Hashes a dictionary using SHAKE-256 and returns the hexdigest.

    Parameters
    ----------
    d : dict
        The dictionary to hash.
    length : int, optional
        The length of the hexdigest in number of text characters, by default 8.

    Returns
    -------
    hexdigest : str
        A string representing the hash in hexadecimal format.
    """
    d_enc = json.dumps(d, sort_keys=True).encode()
    return hashlib.shake_256(d_enc).hexdigest(length // 2)


def hash_function(func, length: int = 8) -> str:
    """Hashes a function using SHAKE-256 and returns the hexdigest.

    Parameters
    ----------
    func : Callable
        The function to hash. Can be a partial function.
    length : int, optional
        The length of the hexdigest in number of text characters, by default 8.

    Returns
    -------
    hexdigest : str
        A string representing the hash in hexadecimal format.
    """
    if not callable(func):
        raise ValueError("The input must be a callable function.")
    elif isinstance(func, partial):
        return hash_function(func.func)

    func_str = func.__code__.co_code
    return hashlib.shake_256(func_str).hexdigest(length // 2)


def standardize_path(path: str | Path) -> str:
    """Represents a posix path as a standardized string."""
    return Path(path).expanduser().resolve().as_posix()


@contextmanager
def suppress_logging(new_level):
    """Suppresses all logs of a given level within a context block."""
    logger = logging.getLogger()
    previous_level = logger.level
    logger.setLevel(new_level)
    try:
        yield
    finally:
        logger.setLevel(previous_level)



#!/usr/bin/env python
import logging
from argparse import ArgumentParser
from collections import defaultdict
from pathlib import Path

from sklearn.inspection import permutation_importance

from folktexts._io import save_json, save_pickle
from folktexts.classifier import LLMClassifier
from folktexts.dataset import Dataset
from folktexts.llm_utils import get_model_folder_path, load_model_tokenizer

# Local paths
# DEFAULT_ROOT_DIR = Path("/fast/groups/sf")            # CLUSTER dir
DEFAULT_ROOT_DIR = Path("~").expanduser().resolve()     # LOCAL dir

DEFAULT_MODELS_DIR = DEFAULT_ROOT_DIR / "huggingface-models"
DEFAULT_DATA_DIR = DEFAULT_ROOT_DIR / "data"
DEFAULT_RESULTS_DIR = Path(".")

DEFAULT_TASK_NAME = "ACSIncome"

DEFAULT_CONTEXT_SIZE = 500
DEFAULT_BATCH_SIZE = 30
DEFAULT_SEED = 42

DEFAULT_SUBSAMPLING = 0.1           # NOTE: by default, uses 10% of the dataset
DEFAULT_PERMUTATION_REPEATS = 5


def setup_arg_parser() -> ArgumentParser:

    # Init parser
    parser = ArgumentParser(description="Evaluate LLM feature importance on a given ACS task.")

    # List of command-line arguments, with type and helper string
    cli_args = [
        ("--model",         str, "[str] Model name or path to model saved on disk"),
        ("--task",          str, "[str] Name of the ACS task to run the experiment on", False, DEFAULT_TASK_NAME),
        ("--results-dir",   str, "[str] Directory under which this experiment's results will be saved", False, DEFAULT_RESULTS_DIR),
        ("--data-dir",      str, "[str] Root folder to find datasets on", False, DEFAULT_DATA_DIR),
        ("--models-dir",    str, "[str] Root folder to find huggingface models on", False, DEFAULT_MODELS_DIR),
        ("--scorer",        str, "[str] Name of the scorer to use for evaluation", False, "roc_auc"),
        ("--batch-size",    int, "[int] The batch size to use for inference", False, DEFAULT_BATCH_SIZE),
        ("--context-size",  int, "[int] The maximum context size when prompting the LLM", False, DEFAULT_CONTEXT_SIZE),
        ("--subsampling",   float, "[float] Which fraction of the dataset to use (if omitted will use all data)", DEFAULT_SUBSAMPLING),
        ("--fit-threshold", int, "[int] Whether to fit the prediction threshold, and on how many samples", False),
        ("--seed",          int, "[int] Random seed -- to set for reproducibility", False, DEFAULT_SEED),
    ]

    for arg in cli_args:
        parser.add_argument(
            arg[0],
            type=arg[1],
            help=arg[2],
            required=(arg[3] if len(arg) > 3 else True),    # NOTE: required by default
            default=(arg[4] if len(arg) > 4 else None),     # default value if provided
        )
    return parser


def parse_feature_importance(results: dict, columns: list[str]) -> dict:
    """Parse the results dictionary of sklearn's permutation_importance."""
    parsed_r = defaultdict(dict)
    for idx, col in enumerate(columns):
        parsed_r[col]["imp_mean"] = results.importances_mean[idx]
        parsed_r[col]["imp_std"] = results.importances_std[idx]

    return parsed_r


def compute_feature_importance(
    llm_clf: LLMClassifier,
    dataset: Dataset,
    scorer: str,
    results_dir: Path,
    fit_threshold=None,
    seed=DEFAULT_SEED,
) -> dict:

    # Get train and test data
    X_test, y_test = dataset.get_test()
    logging.info(f"{X_test.shape=}")

    permutation_kwargs = dict(
        X=X_test, y=y_test,
        scoring=scorer,
        n_repeats=DEFAULT_PERMUTATION_REPEATS,
        random_state=seed,
    )

    # Optionally, fit the LLM classifier's threshold on a few data samples.
    if fit_threshold and isinstance(fit_threshold, int):
        X_train_sample, y_train_sample = dataset.sample_n_train_examples(n=fit_threshold)
        llm_clf.fit(X_train_sample, y_train_sample)

    # LLM feature importance
    r = permutation_importance(llm_clf, **permutation_kwargs)
    llm_imp_file_path = results_dir / f"feature-importance.{llm_clf.task.name}.{llm_clf.model_name}.pkl"
    save_pickle(obj=r, path=llm_imp_file_path)
    save_json(
        parse_feature_importance(results=r, columns=X_test.columns),
        path=llm_imp_file_path.with_suffix(".json"))

    print("LLM feature importance:")
    for i in r.importances_mean.argsort()[::-1]:
        print(
            f"{X_test.columns[i]:<8}"
            f"{r.importances_mean[i]:.3f}"
            f" +/- {r.importances_std[i]:.3f}")

    print(X_test.columns.tolist())


def main():
    # Parse arguments from command line
    args = setup_arg_parser().parse_args()

    # Set logging level
    logging.getLogger().setLevel(logging.INFO)

    # Resolve model folder path
    model_folder_path = Path(get_model_folder_path(model_name=args.model, root_dir=args.models_dir))
    if not model_folder_path.exists() or not model_folder_path.is_dir():
        model_folder_path = Path(args.model).resolve()

    # Load model and tokenizer
    logging.info(f"Loading model from {model_folder_path.as_posix()}")
    model, tokenizer = load_model_tokenizer(model_folder_path)

    # Create results directory if needed
    # Set-up results directory
    from folktexts.cli._utils import get_or_create_results_dir
    results_dir = get_or_create_results_dir(
        model_name=Path(args.model).name,
        task_name=args.task,
        results_root_dir=args.results_dir,
    )
    logging.info(f"Saving results to {results_dir.as_posix()}")

    # Load Task and Dataset
    from folktexts.acs import ACSTaskMetadata
    task = ACSTaskMetadata.get_task(args.task)

    from folktexts.acs import ACSDataset
    dataset = ACSDataset.make_from_task(
        task=task,
        cache_dir=args.data_dir,
        seed=args.seed,
    )

    # Optionally, subsample dataset
    if args.subsampling:
        dataset.subsample(args.subsampling)      # subsample in-place
        logging.info(f"{dataset.subsampling=}")

    # Construct LLM Classifier
    from folktexts.classifier import LLMClassifier
    llm_clf = LLMClassifier(
        model=model,
        tokenizer=tokenizer,
        task=task,
        batch_size=args.batch_size,
        context_size=args.context_size,
    )

    # Compute feature importance
    compute_feature_importance(
        llm_clf,
        dataset=dataset,
        scorer=args.scorer,
        results_dir=results_dir,
        fit_threshold=args.fit_threshold,
        seed=args.seed,
    )


if __name__ == "__main__":
    main()



"""Module to access ACS data using the folktables package.
"""
from __future__ import annotations

import logging
from pathlib import Path

import pandas as pd
from folktables import ACSDataSource
from folktables.load_acs import state_list

from ..dataset import Dataset
from .acs_tasks import ACSTaskMetadata

DEFAULT_DATA_DIR = Path("~/data").expanduser().resolve()
DEFAULT_TEST_SIZE = 0.1
DEFAULT_VAL_SIZE = 0.1
DEFAULT_SEED = 42

DEFAULT_SURVEY_YEAR = "2018"
DEFAULT_SURVEY_HORIZON = "1-Year"
DEFAULT_SURVEY_UNIT = "person"


class ACSDataset(Dataset):
    """Wrapper for ACS folktables datasets."""

    def __init__(
        self,
        data: pd.DataFrame,
        full_acs_data: pd.DataFrame,
        task: ACSTaskMetadata,
        test_size: float = DEFAULT_TEST_SIZE,
        val_size: float = DEFAULT_VAL_SIZE,
        subsampling: float = None,
        seed: int = 42,
    ):
        self._full_acs_data = full_acs_data
        super().__init__(
            data=data,
            task=task,
            test_size=test_size,
            val_size=val_size,
            subsampling=subsampling,
            seed=seed,
        )

    @classmethod
    def make_from_task(
        cls,
        task: str | ACSTaskMetadata,
        cache_dir: str | Path = None,
        survey_year: str = DEFAULT_SURVEY_YEAR,
        horizon: str = DEFAULT_SURVEY_HORIZON,
        survey: str = DEFAULT_SURVEY_UNIT,
        seed: int = DEFAULT_SEED,
        **kwargs,
    ):
        """Construct an ACSDataset object using ACS survey parameters.

        Parameters
        ----------
        task : str | ACSTaskMetadata
            The name of the ACS task or the task object itself.
        cache_dir : str | Path, optional
            The directory where ACS data is (or will be) saved to, by default
            uses DEFAULT_DATA_DIR.
        survey_year : str, optional
            The year from which to load survey data, by default DEFAULT_SURVEY_YEAR.
        horizon : str, optional
            The time horizon of survey data to load, by default DEFAULT_SURVEY_HORIZON.
        survey : str, optional
            The name of the survey unit to load, by default DEFAULT_SURVEY_UNIT.
        seed : int, optional
            The random seed, by default DEFAULT_SEED.
        **kwargs
            Extra key-word arguments to be passed to the Dataset constructor.
        """
        # Create "folktables" sub-folder under the given cache dir
        cache_dir = Path(cache_dir or DEFAULT_DATA_DIR).expanduser().resolve() / "folktables"
        if not cache_dir.exists():
            logging.warning(f"Creating cache directory '{cache_dir}' for ACS data.")
            cache_dir.mkdir(exist_ok=True, parents=False)

        # Parse task if given a string
        task_obj = ACSTaskMetadata.get_task(task) if isinstance(task, str) else task

        # Load ACS data source
        print("Loading ACS data...")
        data_source = ACSDataSource(
            survey_year=survey_year, horizon=horizon, survey=survey,
            root_dir=cache_dir.as_posix(),
        )

        # Get full ACS dataset
        full_acs_data = data_source.get_data(
            states=state_list, download=True, random_seed=seed)

        # Parse data for this task
        parsed_data = cls._parse_task_data(full_acs_data, task_obj)

        return cls(
            data=parsed_data,
            full_acs_data=full_acs_data,
            task=task_obj,
            seed=seed,
            **kwargs,
        )

    @property
    def task(self) -> ACSTaskMetadata:
        return self._task

    @task.setter
    def task(self, new_task: ACSTaskMetadata):
        # Parse data rows for new ACS task
        self._data = self._parse_task_data(self._full_acs_data, new_task)

        # Re-make train/test/val split
        self._train_indices, self._test_indices, self._val_indices = (
            self._make_train_test_val_split(
                self._data, self.test_size, self.val_size, self._rng)
        )

        # Check if sub-sampling is necessary (it's applied only to train/test/val indices)
        if self.subsampling is not None:
            self._subsample_train_test_val_indices(self.subsampling)

        self._task = new_task

    @classmethod
    def _parse_task_data(cls, full_df: pd.DataFrame, task: ACSTaskMetadata) -> pd.DataFrame:
        """Parse a DataFrame for compatibility with the given task object.

        Parameters
        ----------
        full_df : pd.DataFrame
            Full DataFrame. Some rows and/or columns may be discarded for each
            task.
        task : ACSTaskMetadata
            The task object used to parse the given data.

        Returns
        -------
        parsed_df : pd.DataFrame
            Parsed DataFrame in accordance with the given task.
        """
        if not isinstance(task, ACSTaskMetadata):
            logging.error(f"Expected task of type `ACSTaskMetadata` for {type(task)}")
            return full_df

        # Parse data
        parsed_df = task.folktables_obj._preprocess(full_df)

        # Threshold the target column if necessary
        if task.target_threshold is not None and task.get_target() not in parsed_df.columns:
            parsed_df[task.get_target()] = task.target_threshold.apply_to_column_data(parsed_df[task.target])

        return parsed_df



"""Helper function for defining binarization thresholds.
"""
from __future__ import annotations

import dataclasses
import operator
from typing import ClassVar

import pandas as pd


@dataclasses.dataclass(frozen=True, eq=True)
class Threshold:
    """A class to represent a threshold value and its comparison operator.

    Attributes
    ----------
    value : float | int
        The threshold value to compare against.
    op : str
        The comparison operator to use. One of '>', '<', '>=', '<=', '=='.
    """
    value: float | int
    op: str

    valid_ops: ClassVar[dict] = {
        ">": operator.gt,
        "<": operator.lt,
        ">=": operator.ge,
        "<=": operator.le,
        "==": operator.eq,
    }

    def __post_init__(self):
        if self.op not in self.valid_ops.keys():
            raise ValueError(f"Invalid comparison operator '{self.op}'.")

    def __str__(self):
        return f"{self.op}{self.value}"

    def apply_to_column_data(self, data: float | int | pd.Series) -> int | pd.Series:
        """Applies the threshold operation to a pandas Series or scalar value."""
        if isinstance(data, pd.Series):
            return self.valid_ops[self.op](data, self.value).astype(int)
        elif isinstance(data, (float, int)):
            return int(self.valid_ops[self.op](data, self.value))
        else:
            raise TypeError(f"Invalid data type '{type(data)}'.")

    def apply_to_column_name(self, column_name: str) -> str:
        """Standardizes naming of thresholded columns."""
        return column_name + str(self)



#!/usr/bin/env python3
"""Launch htcondor jobs for all ACS benchmark experiments.
"""
import argparse
import logging
import math
from pathlib import Path
from pprint import pprint

from folktexts._io import load_json, save_json
from folktexts.llm_utils import get_model_folder_path, get_model_size_B

from .experiments import Experiment, launch_experiment_job

# All ACS prediction tasks
ACS_TASKS = (
    "ACSIncome",
    "ACSEmployment",
    "ACSMobility",
    "ACSTravelTime",
    "ACSPublicCoverage",
)

################
# Useful paths #
################
ROOT_DIR = Path("/fast/groups/sf")
# ROOT_DIR = Path("~").expanduser().resolve()               # on local machine

# ACS data directory
ACS_DATA_DIR = ROOT_DIR / "data"

# Models save directory
MODELS_DIR = ROOT_DIR / "huggingface-models"


##################
# Global configs #
##################
BATCH_SIZE = 20
CONTEXT_SIZE = 600

JOB_CPUS = 4
JOB_MEMORY_GB = 60
JOB_BID = 250

# LLMs to evaluate
LLM_MODELS = [
    # Google Gemma2 models
    "google/gemma-2b",
    "google/gemma-1.1-2b-it",
    "google/gemma-7b",
    "google/gemma-1.1-7b-it",

    "google/gemma-2-9b",
    "google/gemma-2-9b-it",
    "google/gemma-2-27b",
    "google/gemma-2-27b-it",

    # Meta Llama3 models
    "meta-llama/Meta-Llama-3-8B",
    "meta-llama/Meta-Llama-3-8B-Instruct",
    "meta-llama/Meta-Llama-3-70B",
    "meta-llama/Meta-Llama-3-70B-Instruct",

    # Mistral AI models
    "mistralai/Mistral-7B-v0.1",
    "mistralai/Mistral-7B-Instruct-v0.2",
    "mistralai/Mixtral-8x7B-v0.1",
    "mistralai/Mixtral-8x7B-Instruct-v0.1",
    "mistralai/Mixtral-8x22B-v0.1",
    "mistralai/Mixtral-8x22B-Instruct-v0.1",

    # Yi models
    "01-ai/Yi-34B",
    "01-ai/Yi-34B-Chat",

    # Qwen2 models
    # "Qwen/Qwen2-1.5B",
    # "Qwen/Qwen2-1.5B-Instruct",
    # "Qwen/Qwen2-7B",
    # "Qwen/Qwen2-7B-Instruct",
    # "Qwen/Qwen2-72B",
    # "Qwen/Qwen2-72B-Instruct",
]


# Function that defines common settings among all LLM-as-clf experiments
def make_llm_clf_experiment(
    executable_path: str,
    model_name: str,
    task: str,
    results_dir: str,
    **kwargs,
) -> Experiment:
    """Create an experiment object to run.
    """
    # Get model size
    model_size_B = get_model_size_B(model_name)

    # Get model path
    model_path = get_model_folder_path(model_name, root_dir=MODELS_DIR)
    assert Path(model_path).exists(), f"Model path '{model_path}' does not exist."

    # Split experiment and job kwargs
    job_kwargs = {key: val for key, val in kwargs.items() if key.startswith("job_")}
    experiment_kwargs = {key: val for key, val in kwargs.items() if key not in job_kwargs}

    # Set default job kwargs
    job_kwargs.setdefault("job_cpus", JOB_CPUS)
    job_kwargs.setdefault("job_gpus", math.ceil(model_size_B / 40))     # One GPU per 40B parameters
    job_kwargs.setdefault("job_memory_gb", JOB_MEMORY_GB)
    job_kwargs.setdefault("job_gpu_memory_gb", 35 if model_size_B < 5 else 60)
    job_kwargs.setdefault("job_bid", JOB_BID)

    # Set default experiment kwargs
    n_shots = int(experiment_kwargs.get("few_shot", 1))
    experiment_kwargs.setdefault("batch_size", math.ceil(BATCH_SIZE / n_shots))
    experiment_kwargs.setdefault("context_size", CONTEXT_SIZE * n_shots)
    experiment_kwargs.setdefault("data_dir", ACS_DATA_DIR.as_posix())
    # experiment_kwargs.setdefault("fit_threshold", FIT_THRESHOLD)

    # Define experiment
    exp = Experiment(
        executable_path=executable_path,
        kwargs=dict(
            model=model_path,
            task=task,
            results_dir=results_dir,
            **experiment_kwargs,
        ),
        **job_kwargs,
    )

    # Create LLM results directory
    save_json(
        obj=exp.to_dict(),
        path=Path(results_dir) / f"experiment.{exp.hash()}.json",
        overwrite=True,
    )

    return exp


def setup_arg_parser() -> argparse.ArgumentParser:
    # Init parser
    parser = argparse.ArgumentParser(description="Launch experiments to evaluate LLMs as classifiers.")

    parser.add_argument(
        "--executable-path",
        type=str,
        help="[string] Path to the executable script to run.",
        required=True,
    )

    parser.add_argument(
        "--results-dir",
        type=str,
        help="[string] Directory under which results will be saved.",
        required=True,
    )

    parser.add_argument(
        "--model",
        type=str,
        help="[string] Model name on huggingface hub - can provide multiple!",
        required=False,
        action="append",
    )

    parser.add_argument(
        "--task",
        type=str,
        help="[string] ACS task name to run experiments on - can provide multiple!",
        required=False,
        action="append",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Construct folder structure and print experiments without launching them.",
        default=False,
    )

    parser.add_argument(
        "--experiment-json",
        type=str,
        help="[string] Path to an experiment JSON file to load. Will override all other args.",
        required=False,
    )

    return parser


def main():
    # Parse command-line arguments
    parser = setup_arg_parser()
    args, extra_kwargs = parser.parse_known_args()

    # Parse extra kwargs
    from ._utils import cmd_line_args_to_kwargs
    extra_kwargs = cmd_line_args_to_kwargs(extra_kwargs)

    # Prepare command-line arguments
    models = args.model or LLM_MODELS
    tasks = args.task or ACS_TASKS
    executable_path = Path(args.executable_path).resolve()
    if not executable_path.exists() or not executable_path.is_file():
        raise FileNotFoundError(f"Executable script not found at '{executable_path}'.")

    # Set-up results directory
    Path(args.results_dir).mkdir(parents=True, exist_ok=True)

    # Load experiment from JSON file if provided
    if args.experiment_json:
        print(f"Launching job for experiment at '{args.experiment_json}'...")
        exp = Experiment(**load_json(args.experiment_json))
        all_experiments = [exp]

    # Otherwise, run all experiments planned
    else:
        all_experiments = [
            make_llm_clf_experiment(
                executable_path=executable_path.as_posix(),
                model_name=model,
                task=task,
                results_dir=args.results_dir,
                **extra_kwargs,
            )
            for model in models
            for task in tasks
        ]

    # Log total number of experiments
    print(f"Launching {len(all_experiments)} experiment(s)...\n")
    for i, exp in enumerate(all_experiments):
        cluster_id = launch_experiment_job(exp).cluster() if not args.dry_run else None
        print(f"{i:2}. cluster-id={cluster_id}")
        pprint(exp.to_dict(), indent=4)


if __name__ == "__main__":
    main()



from ._version import __version__, __version_info__
from .acs import ACSDataset, ACSTaskMetadata
from .benchmark import BenchmarkConfig, CalibrationBenchmark
from .classifier import LLMClassifier
from .task import TaskMetadata
